using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using AssetBundles;
using Battlehub.RTCommon;
using SimpleJSON;
using Stopper;
using UnityEngine;
using UnityEngine.UI;
using MVR.FileManagementSecure;
using Battlehub.RTSaveLoad;
using Battlehub.RTCommon;
using UnityStandardAssets.ImageEffects;
/// <summary>

/// </summary>

namespace Stopper {
	public class CUAControllers : MVRScript
	{
        string savePath = "Saves\\CUAController";
		string fileExt = "cuacontroller";

		// CUA Loader
		CustomUnityAssetLoader loader;

		// Prefabricated UI for joint controllers. Not sure how to get ahold of this
		// other than copying from somewhere
		GameObject controllerUIPrefab;
		
		// A sample controller for retrieving controller meshes. Again, not sure
		// how to find this other than copying from elsewhere
		FreeControllerV3 sampleController;
		
		// GameObjects to organize all of our stuff
		GameObject unityAssetControls;
		GameObject controlUI;
		GameObject controllers;

		Atom controlAtom;
		string controlAtomUid;
		Transform cuaRoot;
		Transform root;

		List<FreeControllerV3> createdControllers = new List<FreeControllerV3>();
		List<ConfigurableJoint> controllerJoints = new List<ConfigurableJoint>();

		List<ConfigurableJoint> createdJoints = new List<ConfigurableJoint>();
		List<Rigidbody> createdRigidbodies = new List<Rigidbody>();

		List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();

		// UI
		string selectedRootPath;

		JSONStorableBool createRigidBodies;
		JSONStorableBool createJoints;
		JSONStorableBool loadOnStartup;

		bool loaded = false;
		bool isCUA = false;

		JSONClass savedJson;

		float defaultRbMass = 1f;
		float defaultRbDrag = 1f;
		float defaultJointDriveSpring = 100f;
		float defaultJointDriveDamper = 0f;
		float defaultJointDriveMaxForce = 100f;
		
		Dictionary<string,FreeControllerV3> rigidBodyControllers = new Dictionary<string, FreeControllerV3>();

		public override void Init()
		{
			rigidBodyControllers = new Dictionary<string, FreeControllerV3>();
			loader = containingAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>(true);
			if(loader == null) {
				isCUA = false;
			} else {
				isCUA = true;
			}

			GameObject windowDisplayControl = GameObject.Find("/SceneAtoms/WindowCamera/reParentObject/DisplayControl");
			if(windowDisplayControl == null){
				SuperController.LogError("Unable to find window display control.");
				return;
			}

			sampleController = windowDisplayControl.GetComponent<FreeControllerV3>();

			if(controllerUIPrefab == null) {
				FindSampleControllerUI();
			}
			
			if(controllerUIPrefab == null) { 
				var personAtom = LoadAssetBundleSync("a_per", "Person");
				FindSampleControllerUI();
			}

			if(controllerUIPrefab == null || sampleController == null) {
				SuperController.LogError("Unable to load required templates. Fix this by adding a Person to the scene");
				return;
			}

			JSONStorableStringChooser rootChooser = new JSONStorableStringChooser(
				"Root Node", 
				GetRootCandidates(),
				null, 
				"Root Node", 
				(string path) => {
					selectedRootPath = path;
				});
			CreateScrollablePopup(rootChooser, false);
			RegisterStringChooser(rootChooser);
			

			createRigidBodies = new JSONStorableBool("Create Rigidbodies", true);
			createJoints = new JSONStorableBool("Create Joints", true);
	
			var createRigidBodiesToggle = CreateToggle(createRigidBodies, false);
			RegisterBool(createRigidBodies);
			
			var createJointsToggle = CreateToggle(createJoints, false);
			RegisterBool(createJoints);

			loadOnStartup = new JSONStorableBool("LoadOnStartup", false);
			RegisterBool(loadOnStartup);
			
			UIDynamicButton goButton = CreateButton("Create Controllers", false);
			goButton.button.onClick.AddListener(() => {
				CreateControlAtom();
				AddControllers();
			});

			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});

			AddSlider("Rigidbody Drag", defaultRbDrag, 0f, 2f, (float val) => {
				foreach(var rb in createdRigidbodies) {
					rb.drag = val;
				}
			}, true, true);

			AddSlider("Rigidbody Mass", defaultRbMass, 0f, 2f, (float val) => {
				foreach(var rb in createdRigidbodies) {
					rb.mass = val;
				}
			}, true, true);

			AddSlider("Joint Spring", defaultJointDriveSpring, 0f, 1000f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionSpring = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Damper", defaultJointDriveDamper, 0f, 10f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionDamper = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Max Force", defaultJointDriveMaxForce, 0f, 1000f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.maximumForce = val;
					joint.slerpDrive = drive;
				}
			}, true, true);



			if(isCUA) {
				loader.RegisterAssetLoadedCallback(AssetLoadedCallback);
				loader.RegisterAssetClearedCallback(AssetClearedCallback);
			}
		}

		private List<String> GetRootCandidates(){
			List<string> candidates = new List<string>();

			Queue<Transform> frontier = new Queue<Transform>();

			var asset = containingAtom.GetStorableByID("asset");
			if(asset != null) {
				cuaRoot = asset.transform;
			}
			if(cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject.Find("object/rescaleObject");
			}
			if(cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject;
			}

			frontier.Enqueue(cuaRoot);

			Transform node;
			while(frontier.Count > 0){
				node = frontier.Dequeue();
				if(frontier == null)
					break;
				for(int i = 0; i < node.childCount; i++){
					var child = node.GetChild(i);
					if(child.childCount == 0) {
						continue;
					}
					if(candidates.Count < 30){
						candidates.Add(GetPathToRoot(child, cuaRoot));
						frontier.Enqueue(child);
					} else {
						return candidates;
					}
				}
			}

			return candidates;
		}

		private string GetPathToRoot(Transform t, Transform root){
			if (t.parent == null || t.parent == root ){
				return t.name;
			} else {
				return GetPathToRoot(t.parent, root) + "/" + t.name;
			}
		}

		private GameObject LoadAssetBundleSync(string assetBundleName, string assetName, bool deactivate = true) {
			GameObject asset = null;
			
			string error;
			var loadedAssetBundle = AssetBundleManager.GetLoadedAssetBundle(assetBundleName, out error);

			if(loadedAssetBundle == null) {
				AssetBundleLoader assetLoader = new AssetBundleLoader();
				assetLoader.Load(assetBundleName, (bundleName, assetBundle) => {
					var loadedAsset = assetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject;
					
					if(deactivate) {
						loadedAsset.SetActive(false);
					}

					asset = UnityEngine.Object.Instantiate<GameObject>(loadedAsset);

					if(deactivate) {
						loadedAsset.SetActive(true);
					}
					
					assetBundle.Unload(true);
				});
			} else {
				var loadedAsset = loadedAssetBundle.m_AssetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject;
				if(deactivate) {
					loadedAsset.SetActive(false);
				}
				asset = UnityEngine.Object.Instantiate<GameObject>(loadedAsset);
				if(deactivate) {
					loadedAsset.SetActive(true);
				}
			}
			
			return asset;
		}

		private void CreateControlAtom() {
			
			// Initialize base objects
			
			// The top-level Atom is already initialized and has some internal state baked in.
			// We need to add the controllers to a new atom to make selection and timeline work
			controlAtomUid = $"{containingAtom.uid}#Controllers";

			if(controlAtom == null) {
				// SuperController.LogMessage($"CreateControlAtom {this.containingAtom}");
				var existingControlAtom = SuperController.singleton.GetAtomByUid(controlAtomUid);

				if(existingControlAtom != null) {
					//SuperController.LogMessage($"Remove existing {this.containingAtom}");
					SuperController.singleton.RemoveAtom(existingControlAtom, true);
				} 

				// Get new copy of the 'empty' prefab.
				GameObject instantiatedEmpty = LoadAssetBundleSync("a_mis", "Empty");
				controlAtom = instantiatedEmpty.GetComponent<Atom>();

				unityAssetControls = CreateChildObject(controlAtom.reParentObject.gameObject, "CustomUnityAssetControls");
				//controlUI = controlAtom.transform.Find("UIPlaceHolder").gameObject;
				controlUI = CreateChildObject(unityAssetControls, "UI");
				controllers = CreateChildObject(unityAssetControls, "Controllers");
			}
		}

		private void InitControlAtom() {
			if(controlAtom != null && !controlAtom.gameObject.active) {
				controlAtom.gameObject.SetActive(true);
				controlAtom.parentAtom = containingAtom;
				//SuperController.LogMessage($"InitControlAtom {this.containingAtom}");
				SuperController.singleton.AddAtom(controlAtom, controlAtomUid, false, false, false, false);
			}
		}

		private void AddControllers() {
			//SuperController.LogMessage($"Adding Controllers {this.containingAtom}");
			if(loaded) {
				SuperController.LogMessage("Already loaded. Reload plugin if needed.");
				return;
			}

			if(selectedRootPath == null){
				SuperController.LogMessage("No root node selected.");
				return;
			}

			root = cuaRoot.Find(selectedRootPath);

			if(root == null){
				SuperController.LogMessage($"Could not find root node at path {selectedRootPath}");
				return;
			}

			ProcessRigging(root,0);

			if(createJoints.val){
				ConfigurableJoint jointToRootControl = CreateComponent(root.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
				jointToRootControl.connectedBody = containingAtom.mainController.GetComponent<Rigidbody>();

				jointToRootControl.xMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.yMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.zMotion = ConfigurableJointMotion.Locked;
				
				jointToRootControl.angularXMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.angularYMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.angularZMotion = ConfigurableJointMotion.Locked;
				JointDrive drive = new JointDrive();
				drive.positionSpring = 100000f;
				drive.positionDamper = 0f;
				drive.maximumForce = 10000f;
				jointToRootControl.xDrive = drive;
				jointToRootControl.yDrive = drive;
				jointToRootControl.zDrive = drive;
				jointToRootControl.angularXDrive = drive;
				jointToRootControl.angularYZDrive = drive;
			}
			
	
			foreach(Rigidbody rb in root.GetComponentsInChildren<Rigidbody>(true)){
				AddController(rb.name);
			}

			// If the control atom hasn't been initialized, do it now;
			InitControlAtom();

			foreach(Rigidbody rb in root.GetComponentsInChildren<Rigidbody>(true)){
				LinkRBToController(rb);
			}

			RestoreControlAtom(savedJson);
			LateRestoreControlAtom(savedJson);

			loaded = true;
			loadOnStartup.val = true;
		}

		private void AssetLoadedCallback() {
			// If the control atom hasn't been created, do it now
			CreateControlAtom();
			AddControllers();
		}

		private void AssetClearedCallback() {
			OnDestroy();
			// OnDestroy will deregister this
			if(isCUA) {
				loader.RegisterAssetLoadedCallback(AssetLoadedCallback);
			}
		}

		private void FindSampleControllerUI() {
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name.Contains("UIFreeControllerJoint")) {
					controllerUIPrefab = go;
					break;
				}
			}
		}
		
		private Rigidbody ProcessRigging(Transform parent, int depth){
			Rigidbody parentRB = parent.gameObject.GetComponent<Rigidbody>();
			if(parentRB == null && createRigidBodies.val){
				parentRB = CreateComponent(parent.gameObject, typeof(Rigidbody)) as Rigidbody;
				parentRB.isKinematic = false;
				parentRB.useGravity = false;
				parentRB.detectCollisions = true;
				parentRB.drag = defaultRbDrag;
				parentRB.mass = defaultRbMass;
				createdRigidbodies.Add(parentRB);
			}

			try{
			//	SuperController.LogMessage($"Processing: {parent}");
				for( int i = 0; i < parent.childCount; i++) {
					Transform child = parent.GetChild(i);
			//		SuperController.LogMessage($"Child: {child}");

					Rigidbody childRB = child.GetComponent<Rigidbody>();
					if(childRB == null){
						if(createRigidBodies.val){
					 		childRB = CreateComponent(child.gameObject, typeof(Rigidbody)) as Rigidbody;
							childRB.isKinematic = false;
							childRB.useGravity = false;
							childRB.detectCollisions = true;
							childRB.drag = defaultRbDrag;
							childRB.mass = defaultRbMass;
							createdRigidbodies.Add(childRB);
						} else {
							continue;
						}
					}
					
					if(parentRB !=null && createJoints.val){
							
						ConfigurableJoint childJoint = CreateComponent(child.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
						childJoint.connectedBody = parentRB;
						createdJoints.Add(childJoint);
						
						childJoint.xMotion = ConfigurableJointMotion.Locked;
						childJoint.yMotion = ConfigurableJointMotion.Locked;
						childJoint.zMotion = ConfigurableJointMotion.Locked;
						
						childJoint.angularXMotion = ConfigurableJointMotion.Limited;
						childJoint.angularYMotion = ConfigurableJointMotion.Limited;
						childJoint.angularZMotion = ConfigurableJointMotion.Limited;

						childJoint.rotationDriveMode = RotationDriveMode.Slerp;
						JointDrive slerpDrive = new JointDrive();
						slerpDrive.positionSpring = defaultJointDriveSpring;
						slerpDrive.positionDamper = defaultJointDriveDamper;
						slerpDrive.maximumForce = defaultJointDriveMaxForce;
						childJoint.slerpDrive = slerpDrive;
						
						SoftJointLimit xLowJointLimit = new SoftJointLimit();
						xLowJointLimit.limit = -180f;
						SoftJointLimit jointLimit = new SoftJointLimit();
						jointLimit.limit = 180f;
						childJoint.lowAngularXLimit = xLowJointLimit;
						childJoint.highAngularXLimit = jointLimit;
						childJoint.angularYLimit =jointLimit;
						childJoint.angularZLimit = jointLimit;
			
					}
					
					ProcessRigging(child, depth+1);
				}
				
			}
			catch (Exception ex)
			{
				SuperController.LogError($"Failed to init UnityObjectController: {ex}");
			}
			return parentRB;
		}

		private void AddController(string rigidBodyName){
			try {
				if(rigidBodyControllers.ContainsKey(rigidBodyName)) {
					// already exists
					return;
				}

				GameObject rbControl = CreateChildObject(controllers, $"{rigidBodyName}Control");
				rbControl.SetActive(false);
				rbControl.layer = 8;
				
				Rigidbody controlRB = CreateComponent(rbControl, typeof(Rigidbody)) as Rigidbody;
				controlRB.isKinematic = true;
				controlRB.useGravity = false;
				controlRB.detectCollisions = true;

				/*	
				ConfigurableJoint controlJoint = CreateComponent(rb.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
				controlJoint.connectedBody = controlRB;
				controlJoint.autoConfigureConnectedAnchor = false;
				controlJoint.anchor = Vector3.zero;
				controlJoint.connectedAnchor = Vector3.zero;
				
				if( !controllerJoints.Contains(controlJoint)){
					controllerJoints.Add(controlJoint);
				}
			*/
				FreeControllerV3 controller = CreateComponent(rbControl, typeof(FreeControllerV3)) as FreeControllerV3;
				controller.containingAtom = controlAtom;
			//	controller.followWhenOff = rb.transform;
				controller.useGravityOnRBWhenOff = false;
				controller.meshScale = 0.5f;
				controller.selectedScale = 0.1f;
				controller.enableSelectRoot = false;
				controller.enabled = true;
				controller.hidden = false;
				controller.startingPositionState = FreeControllerV3.PositionState.Off;
				controller.startingRotationState = FreeControllerV3.RotationState.Off;

				if( !createdControllers.Contains(controller)){
					createdControllers.Add(controller);
				}
				rigidBodyControllers[rigidBodyName] = controller;

				// Copy meshes from the sample controller
				controller.material = sampleController.material;
				if(controller.onPositionMesh == null){
					controller.onPositionMesh = sampleController.onPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onPositionMesh, controller.transform);
					controller.offPositionMesh = sampleController.offPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offPositionMesh, controller.transform);
					controller.followingPositionMesh = sampleController.followingPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingPositionMesh, controller.transform);
					controller.holdPositionMesh = sampleController.holdPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdPositionMesh, controller.transform);
					controller.lockPositionMesh = sampleController.lockPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockPositionMesh, controller.transform);
					controller.onRotationMesh = sampleController.onRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onRotationMesh, controller.transform);
					controller.offRotationMesh = sampleController.offRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offRotationMesh, controller.transform);
					controller.followingRotationMesh = sampleController.followingRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingRotationMesh, controller.transform);
					controller.holdRotationMesh = sampleController.holdRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdRotationMesh, controller.transform);
					controller.lockRotationMesh = sampleController.lockRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockRotationMesh, controller.transform);
					controller.lookAtRotationMesh = sampleController.lookAtRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lookAtRotationMesh, controller.transform);
					controller.moveModeOverlayMesh = sampleController.moveModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.moveModeOverlayMesh, controller.transform);
					controller.rotateModeOverlayMesh = sampleController.rotateModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.rotateModeOverlayMesh, controller.transform);
					controller.deselectedMesh = sampleController.deselectedMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.deselectedMesh, controller.transform);
				}
				
				SphereCollider collider = CreateComponent(rbControl, typeof(SphereCollider)) as SphereCollider;
				collider.enabled = true;
				collider.radius = 0.1f;

				// Controller Link. Not sure what this is used for.
				GameObject controlLink = CreateChildObject(rbControl, rigidBodyName+"Link");
				controlLink.layer = 8;
				
				SphereCollider linkCollider = CreateComponent(controlLink, typeof(SphereCollider)) as SphereCollider;
				linkCollider.enabled = true;
				linkCollider.radius = 0.01f;
				
				Rigidbody linkRB = CreateComponent(controlLink, typeof(Rigidbody)) as Rigidbody;
				linkRB.isKinematic = true;
				linkRB.useGravity = false;
				linkRB.detectCollisions = false;

				FreeControllerV3Link freeControllerLink = CreateComponent(controlLink, typeof(FreeControllerV3Link)) as FreeControllerV3Link;
				freeControllerLink.linkedController = controller;
				

				// UI
				GameObject uiPlaceholder = CreateChildObject(controlUI, $"{rigidBodyName}UIPlaceholder" );
				uiPlaceholder.SetActive(false);

				MoveAndRotateAsHUDAnchor moveAsHud = CreateComponent(uiPlaceholder, typeof(MoveAndRotateAsHUDAnchor)) as MoveAndRotateAsHUDAnchor;
				moveAsHud.scaleAsAnchor = true;
				moveAsHud.anchorNum = HUDAnchor.AnchorNum.Two;
									
				UIConnectorMaster connectorMaster = CreateComponent(uiPlaceholder, typeof(UIConnectorMaster)) as UIConnectorMaster;
				connectorMaster.UIPrefab = controllerUIPrefab.transform;
									
				UIConnector connector = CreateComponent(uiPlaceholder, typeof(UIConnector)) as UIConnector;
				connector.receiver = controller;
				connector.receiverTransform = rbControl.transform;
				
				controller.UITransforms = new Transform[]{uiPlaceholder.transform};
				controller.UITransformsPlayMode = new Transform[]{};
				
				rbControl.SetActive(true);
			} catch (Exception e) {
				SuperController.LogMessage($"{e}");
			}
		}

		public void LinkRBToController(Rigidbody rb) {
			if(!rigidBodyControllers.ContainsKey(rb.name)){
				SuperController.LogError($"Controller does not exist for rigidbody {rb.name}");
				return;
			}

			FreeControllerV3 controller = rigidBodyControllers[rb.name];
			Rigidbody controlRB = controller.gameObject.GetComponent<Rigidbody>();
			//SuperController.LogMessage($"Linking ${rb} to {controlRB}");
			if(controller.followWhenOff != null) {
				SuperController.LogMessage($"Controller {rb.name} already linked");
				return;
			}
			var pos = controlRB.transform.position;
			var rot = controlRB.transform.rotation;

			controlRB.transform.position = rb.transform.position;
			controlRB.transform.rotation = rb.transform.rotation;

			ConfigurableJoint controlJoint = CreateComponent(rb.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
			controlJoint.connectedBody = controlRB;
			controlJoint.autoConfigureConnectedAnchor = false;
			controlJoint.anchor = Vector3.zero;
			controlJoint.connectedAnchor = Vector3.zero;
			
			if( !controllerJoints.Contains(controlJoint)){
				controllerJoints.Add(controlJoint);
			}

			controller.followWhenOffRB = rb;

			controlRB.transform.position = pos;
			controlRB.transform.rotation = rot;
		}

		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = true, bool rightSide = false){
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = rangeAdjust;
			uiSlider.valueFormat = "F3";
			RegisterFloat(storableFloat);
			
			if(onChange != null){
				storableFloat.setCallbackFunction = onChange;
			}
			
			return storableFloat;
		}

		private IEnumerator WaitForRootLoaded() {
			root = cuaRoot.Find(selectedRootPath);
			while(root == null) {
				yield return null;
				root = cuaRoot.Find(selectedRootPath);
				
			}
			AddControllers();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

			JSONArray jsonArray = new JSONArray();
			controlAtom.Store(jsonArray,true,true);
			
			if(jsonArray.Count > 0) {
			//	SuperController.LogMessage($"storing {jsonArray[0]}");
				jc["controllerAtom"] = jsonArray[0];
			}
			jc["mainController"] = containingAtom.mainController.GetJSON();

			JSONArray rigidbodyNames = new JSONArray();
			foreach(var entry in rigidBodyControllers) {
				rigidbodyNames.Add(entry.Key);
			}
			jc["rigidbodyNames"] = rigidbodyNames;
			return jc;
		}

		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true){
			//SuperController.LogMessage($"RestoreFromJSON {this.containingAtom}");
            base.RestoreFromJSON(jc,restorePhysical,restoreAppearance,presetAtoms,false);
			savedJson = jc;

			CreateControlAtom();

			if(jc["rigidbodyNames"] != null) {
				foreach(var rbJson in jc["rigidbodyNames"].AsArray) {
					AddController((rbJson as JSONNode).Value);
				}
				InitControlAtom();
			}

			RestoreControlAtom(jc);
			if(selectedRootPath != null && !isCUA) {
				StartCoroutine("WaitForRootLoaded");
			}
        }

		public override void PostRestore() {
			foreach(var atom in SuperController.singleton.GetAtoms()) {
				foreach(var controller in atom.freeControllers) {
					atom.RestoreFromLast(controller);
				}
			}
		}

        private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);

			if(loaded) {
				RestoreControlAtom(presetJson);
				LateRestoreControlAtom(presetJson);
			} else {
				AddControllers();
			}
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}

		private void RestoreControlAtom(JSONClass jc) {
			// SuperController.LogMessage($"RestoreControlAtom {this.containingAtom}");
			if(jc == null) {
				return;
			}

			if(jc["mainController"] != null) {
				containingAtom.mainController.RestoreFromJSON(jc["mainController"] as JSONClass);
				var preset = containingAtom.GetStorableByID("Preset");
				preset.transform.position = containingAtom.mainController.transform.position;
				preset.transform.localPosition = containingAtom.mainController.transform.localPosition;
				preset.transform.rotation = containingAtom.mainController.transform.rotation;
			}

			JSONClass controllerAtomJson = jc["controllerAtom"] as JSONClass;
			if(controllerAtomJson != null && controlAtom != null) {
				controlAtom.Restore(controllerAtomJson);
			}
		}

		private void LateRestoreControlAtom(JSONClass jc) {
			// SuperController.LogMessage($"LateRestoreControlAtom {this.containingAtom}");
			if(jc == null) {
				return;
			}

			JSONClass controllerAtomJson = jc["controllerAtom"] as JSONClass;
			if(controllerAtomJson != null && controlAtom != null) {
				controlAtom.LateRestore(controllerAtomJson);
			}
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private GameObject CreateChildObject(GameObject parent, string name) {
			GameObject gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
			createdObjects.Add(gameObject);

			return gameObject;
		}
		
		private Component CreateComponent(GameObject parent, System.Type type) {
			Component component = parent.AddComponent(type) as Component;
			createdObjects.Add(component);
			return component;
		}

		public void OnDestroy()
		{
			//foreach(FreeControllerV3 controller in createdControllers) {
			//	SuperController.singleton.GetAllFreeControllers().Remove(controller);
			//}
			if(loader != null) {
				loader.DeregisterAssetLoadedCallback(AssetLoadedCallback);
				loader.DeregisterAssetClearedCallback(AssetClearedCallback);
			}

			foreach(var o in createdObjects){
				Destroy(o);
			}
			createdObjects.Clear();
			
			if(unityAssetControls != null) {
				Destroy(unityAssetControls);
				unityAssetControls = null;
			}

			if(controlAtom != null){
				SuperController.singleton.RemoveAtom(controlAtom);
				controlAtom = null;
			}
			
		}
	}																
}